# -*- coding: utf-8 -*-

"""

ABGLEICH
zfs sync tool
https://github.com/pleiszenburg/abgleich

    src/abgleich/core/comparison.py: ZFS comparison

    Copyright (C) 2019-2020 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the GNU Lesser General Public License
Version 2.1 ("LGPL" or "License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
https://github.com/pleiszenburg/abgleich/blob/master/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

import itertools
import typing

import typeguard

from .abc import ComparisonABC, ComparisonItemABC, DatasetABC, SnapshotABC, ZpoolABC

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# TYPING
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

ComparisonParentTypes = typing.Union[
    ZpoolABC, DatasetABC, None,
]
ComparisonMergeTypes = typing.Union[
    typing.Generator[DatasetABC, None, None], typing.Generator[SnapshotABC, None, None],
]
ComparisonItemType = typing.Union[
    DatasetABC, SnapshotABC, None,
]
ComparisonStrictItemType = typing.Union[
    DatasetABC, SnapshotABC,
]

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# CLASS
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


@typeguard.typechecked
class Comparison(ComparisonABC):
    def __init__(
        self,
        a: ComparisonParentTypes,
        b: ComparisonParentTypes,
        merged: typing.List[ComparisonItemABC],
    ):

        assert a is not None or b is not None
        if a is not None and b is not None:
            assert type(a) == type(b)

        self._a, self._b, self._merged = a, b, merged

    def __len__(self) -> int:

        return len(self._merged)

    @property
    def a(self) -> ComparisonParentTypes:

        return self._a

    @property
    def a_head(self) -> typing.List[ComparisonStrictItemType]:

        return self._head(
            source=[item.a for item in self._merged],
            target=[item.b for item in self._merged],
        )

    @property
    def a_overlap_tail(self) -> typing.List[ComparisonStrictItemType]:

        return self._overlap_tail(
            source=[item.a for item in self._merged],
            target=[item.b for item in self._merged],
        )

    @property
    def b(self) -> ComparisonParentTypes:

        return self._b

    @property
    def b_head(self) -> typing.List[ComparisonStrictItemType]:

        return self._head(
            source=[item.b for item in self._merged],
            target=[item.a for item in self._merged],
        )

    @property
    def b_overlap_tail(self) -> typing.List[ComparisonStrictItemType]:

        return self._overlap_tail(
            source=[item.b for item in self._merged],
            target=[item.a for item in self._merged],
        )

    @property
    def merged(self) -> typing.Generator[ComparisonItemABC, None, None]:

        return (item for item in self._merged)

    @classmethod
    def _head(
        cls,
        source: typing.List[ComparisonItemType],
        target: typing.List[ComparisonItemType],
    ) -> typing.List[ComparisonItemType]:
        """
        Returns new elements from source.
        If target is empty, returns source.
        If head of target and head of source are identical, returns empty list.
        """

        source, target = cls._strip_none(source), cls._strip_none(target)

        if any((element is None for element in source)):
            raise ValueError("source is not consecutive")
        if any((element is None for element in target)):
            raise ValueError("target is not consecutive")

        if len(source) == 0:
            raise ValueError("source must not be empty")

        if len(set([item.name for item in source])) != len(source):
            raise ValueError("source contains doublicate entires")
        if len(set([item.name for item in target])) != len(target):
            raise ValueError("target contains doublicate entires")

        if len(target) == 0:
            return source  # all of source, target is empty

        try:
            source_index = [item.name for item in source].index(target[-1].name)
        except ValueError:
            raise ValueError("last target element not in source")

        old_source = source[: source_index + 1]

        if len(old_source) <= len(target):
            if target[-len(old_source) :] != old_source:
                raise ValueError(
                    "no clean match between end of target and beginning of source"
                )
        else:
            if target != source[source_index + 1 - len(target) : source_index + 1]:
                raise ValueError(
                    "no clean match between entire target and beginning of source"
                )

        return source[source_index + 1 :]

    @classmethod
    def _overlap_tail(
        cls,
        source: typing.List[ComparisonItemType],
        target: typing.List[ComparisonItemType],
    ) -> typing.List[ComparisonItemType]:
        """
        Overlap must include first element of source.
        """

        source, target = cls._strip_none(source), cls._strip_none(target)

        if len(source) == 0 or len(target) == 0:
            return []

        if any((element is None for element in source)):
            raise ValueError("source is not consecutive")
        if any((element is None for element in target)):
            raise ValueError("target is not consecutive")

        source_names = {item.name for item in source}
        target_names = {item.name for item in target}

        if len(source_names) != len(source):
            raise ValueError("source contains doublicate entires")
        if len(target_names) != len(target):
            raise ValueError("target contains doublicate entires")

        overlap_tail = []
        for item in source:
            if item.name not in target_names:
                break
            overlap_tail.append(item)

        if len(overlap_tail) == 0:
            return overlap_tail

        target_index = target.index(overlap_tail[0])
        if overlap_tail != target[target_index : target_index + len(overlap_tail)]:
            raise ValueError("no clean match in overlap area")

        return overlap_tail

    @classmethod
    def _strip_none(
        cls, elements: typing.List[ComparisonItemType]
    ) -> typing.List[ComparisonItemType]:

        elements = cls._left_strip_none(elements)  # left strip
        elements.reverse()  # flip into reverse
        elements = cls._left_strip_none(elements)  # right strip
        elements.reverse()  # flip back

        return elements

    @staticmethod
    def _left_strip_none(
        elements: typing.List[ComparisonItemType],
    ) -> typing.List[ComparisonItemType]:

        return list(itertools.dropwhile(lambda element: element is None, elements))

    @staticmethod
    def _single_items(
        items_a: typing.Union[ComparisonMergeTypes, None],
        items_b: typing.Union[ComparisonMergeTypes, None],
    ) -> typing.List[ComparisonItemABC]:

        assert items_a is not None or items_b is not None

        if items_a is None:
            return [ComparisonItem(None, item) for item in items_b]
        return [ComparisonItem(item, None) for item in items_a]

    @staticmethod
    def _merge_datasets(
        items_a: typing.Generator[DatasetABC, None, None],
        items_b: typing.Generator[DatasetABC, None, None],
    ) -> typing.List[ComparisonItemABC]:

        items_a = {item.subname: item for item in items_a}
        items_b = {item.subname: item for item in items_b}

        names = list(items_a.keys() | items_b.keys())
        merged = [
            ComparisonItem(items_a.get(name, None), items_b.get(name, None))
            for name in names
        ]
        merged.sort(key=lambda item: item.get_item().name)

        return merged

    @classmethod
    def from_zpools(
        cls,
        zpool_a: typing.Union[ZpoolABC, None],
        zpool_b: typing.Union[ZpoolABC, None],
    ) -> ComparisonABC:

        assert zpool_a is not None or zpool_b is not None

        if zpool_a is None or zpool_b is None:
            return cls(
                a=zpool_a,
                b=zpool_b,
                merged=cls._single_items(
                    getattr(zpool_a, "datasets", None),
                    getattr(zpool_b, "datasets", None),
                ),
            )

        assert zpool_a is not zpool_b
        assert zpool_a != zpool_b

        return cls(
            a=zpool_a,
            b=zpool_b,
            merged=cls._merge_datasets(zpool_a.datasets, zpool_b.datasets),
        )

    @staticmethod
    def _merge_snapshots(
        items_a: typing.Generator[SnapshotABC, None, None],
        items_b: typing.Generator[SnapshotABC, None, None],
    ) -> typing.List[ComparisonItemABC]:

        items_a = list(items_a)
        items_b = list(items_b)
        names_a = [item.name for item in items_a]
        names_b = [item.name for item in items_b]

        assert len(set(names_a)) == len(items_a)  # unique names
        assert len(set(names_b)) == len(items_b)  # unique names

        if len(items_a) == 0 and len(items_b) == 0:
            return []
        if len(items_a) == 0:
            return [ComparisonItem(None, item) for item in items_b]
        if len(items_b) == 0:
            return [ComparisonItem(item, None) for item in items_a]

        try:
            start_b = names_a.index(names_b[0])
        except ValueError:
            start_b = None
        try:
            start_a = names_b.index(names_a[0])
        except ValueError:
            start_a = None

        assert start_a is not None or start_b is not None  # overlap

        prefix_a = [] if start_a is None else [None for _ in range(start_a)]
        prefix_b = [] if start_b is None else [None for _ in range(start_b)]
        items_a = prefix_a + items_a
        items_b = prefix_b + items_b
        suffix_a = (
            []
            if len(items_a) >= len(items_b)
            else [None for _ in range(len(items_b) - len(items_a))]
        )
        suffix_b = (
            []
            if len(items_b) >= len(items_a)
            else [None for _ in range(len(items_a) - len(items_b))]
        )
        items_a = items_a + suffix_a
        items_b = items_b + suffix_b

        assert len(items_a) == len(items_b)

        alt_a, alt_b, state_a, state_b = 0, 0, False, False
        merged = []
        for item_a, item_b in zip(items_a, items_b):
            new_state_a, new_state_b = item_a is not None, item_b is not None
            if new_state_a != state_a:
                alt_a, state_a = alt_a + 1, new_state_a
                if alt_a > 2:
                    raise ValueError("gap in snapshot series")
            if new_state_b != state_b:
                alt_b, state_b = alt_b + 1, new_state_b
                if alt_b > 2:
                    raise ValueError("gap in snapshot series")
            if state_a and state_b:
                if item_a.name != item_b.name:
                    raise ValueError("inconsistent snapshot names")
            merged.append(ComparisonItem(item_a, item_b))

        return merged

    @classmethod
    def from_datasets(
        cls,
        dataset_a: typing.Union[DatasetABC, None],
        dataset_b: typing.Union[DatasetABC, None],
    ) -> ComparisonABC:

        assert dataset_a is not None or dataset_b is not None

        if dataset_a is None or dataset_b is None:
            return cls(
                a=dataset_a,
                b=dataset_b,
                merged=cls._single_items(
                    getattr(dataset_a, "snapshots", None),
                    getattr(dataset_b, "snapshots", None),
                ),
            )

        assert dataset_a is not dataset_b
        assert dataset_a == dataset_b

        return cls(
            a=dataset_a,
            b=dataset_b,
            merged=cls._merge_snapshots(dataset_a.snapshots, dataset_b.snapshots),
        )


@typeguard.typechecked
class ComparisonItem(ComparisonItemABC):
    def __init__(self, a: ComparisonItemType, b: ComparisonItemType):

        assert a is not None or b is not None
        if a is not None and b is not None:
            assert type(a) == type(b)

        self._a, self._b = a, b

    def get_item(self) -> ComparisonStrictItemType:

        if self._a is not None:
            return self._a
        return self._b

    @property
    def complete(self) -> bool:

        return self._a is not None and self._b is not None

    @property
    def a(self) -> ComparisonItemType:

        return self._a

    @property
    def b(self) -> ComparisonItemType:

        return self._b



# -*- coding: utf-8 -*-

"""

ABGLEICH
zfs sync tool
https://github.com/pleiszenburg/abgleich

    src/abgleich/core/snapshot.py: ZFS snapshot

    Copyright (C) 2019-2020 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the GNU Lesser General Public License
Version 2.1 ("LGPL" or "License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
https://github.com/pleiszenburg/abgleich/blob/master/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

import typing

import typeguard

from .abc import ConfigABC, PropertyABC, SnapshotABC, TransactionABC
from .command import Command
from .i18n import t
from .lib import root
from .property import Property
from .transaction import Transaction, TransactionMeta

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# CLASS
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


@typeguard.typechecked
class Snapshot(SnapshotABC):
    def __init__(
        self,
        name: str,
        parent: str,
        properties: typing.Dict[str, PropertyABC],
        context: typing.List[SnapshotABC],
        side: str,
        config: ConfigABC,
    ):

        self._name = name
        self._parent = parent
        self._properties = properties
        self._context = context
        self._side = side
        self._config = config

        self._root = root(config[side]["zpool"], config[side]["prefix"])

        assert self._parent.startswith(self._root)
        self._subparent = self._parent[len(self._root) :].strip("/")

    def __eq__(self, other: SnapshotABC) -> bool:

        return self.subparent == other.subparent and self.name == other.name

    def __getitem__(self, name: str) -> PropertyABC:

        return self._properties[name]

    def get_cleanup_transaction(self) -> TransactionABC:

        assert self._side == "source"

        return Transaction(
            meta=TransactionMeta(
                **{
                    t("type"): t("cleanup_snapshot"),
                    t("snapshot_subparent"): self._subparent,
                    t("snapshot_name"): self._name,
                }
            ),
            commands=[
                Command.on_side(
                    ["zfs", "destroy", f"{self._parent:s}@{self._name:s}"],
                    self._side,
                    self._config,
                )
            ],
        )

    def get_backup_transaction(
        self, source_dataset: str, target_dataset: str,
    ) -> TransactionABC:

        assert self._side == "source"

        ancestor = self.ancestor

        commands = [
            Command.on_side(
                ["zfs", "send", "-c", f"{source_dataset:s}@{self.name:s}",]
                if ancestor is None
                else [
                    "zfs",
                    "send",
                    "-c",
                    "-i",
                    f"{source_dataset:s}@{ancestor.name:s}",
                    f"{source_dataset:s}@{self.name:s}",
                ],
                "source",
                self._config,
            ),
            Command.on_side(
                ["zfs", "receive", f"{target_dataset:s}"], "target", self._config
            ),
        ]

        return Transaction(
            meta=TransactionMeta(
                **{
                    t("type"): t("transfer_snapshot")
                    if ancestor is None
                    else t("transfer_snapshot_incremental"),
                    t("snapshot_subparent"): self._subparent,
                    t("ancestor_name"): "" if ancestor is None else ancestor.name,
                    t("snapshot_name"): self.name,
                }
            ),
            commands=commands,
        )

    @property
    def name(self) -> str:

        return self._name

    @property
    def parent(self) -> str:

        return self._parent

    @property
    def subparent(self) -> str:

        return self._subparent

    @property
    def ancestor(self) -> typing.Union[None, SnapshotABC]:

        assert self in self._context
        self_index = self._context.index(self)

        if self_index == 0:
            return None
        return self._context[self_index - 1]

    @property
    def root(self) -> str:

        return self._root

    @classmethod
    def from_entity(
        cls,
        name: str,
        entity: typing.List[typing.List[str]],
        context: typing.List[SnapshotABC],
        side: str,
        config: ConfigABC,
    ) -> SnapshotABC:

        properties = {
            property.name: property
            for property in (Property.from_params(*params) for params in entity)
        }

        parent, name = name.split("@")

        return cls(
            name=name,
            parent=parent,
            properties=properties,
            context=context,
            side=side,
            config=config,
        )



# -*- coding: utf-8 -*-

"""

ABGLEICH
zfs sync tool
https://github.com/pleiszenburg/abgleich

    src/abgleich/gui/wizard.py: wizard gui

    Copyright (C) 2019-2020 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the GNU Lesser General Public License
Version 2.1 ("LGPL" or "License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
https://github.com/pleiszenburg/abgleich/blob/master/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

import os

from PyQt5.QtGui import QIcon
from PyQt5.QtWidgets import QApplication, QMessageBox
from typeguard import typechecked

from .transaction import TransactionListModel
from .wizard_base import WizardUiBase
from ..core.abc import ConfigABC
from ..core.transaction import TransactionList
from ..core.i18n import t
from ..core.zpool import Zpool
from .. import __version__

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# CLASS
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


@typechecked
class WizardUi(WizardUiBase):
    def __init__(self, config: ConfigABC):

        super().__init__()
        self._config = config

        self.setWindowIcon(
            QIcon(os.path.join(os.path.dirname(__file__), "..", "share", "icon.svg"))
        )
        self.setWindowTitle(f'{t("abgleich wizard"):s} {__version__:s}')

        self._ui["button_cancel"].setEnabled(False)
        self._ui["button_continue"].setEnabled(False)

        self._ui["button_cancel"].clicked.connect(self._cancel_click)
        self._ui["button_continue"].clicked.connect(self._continue_click)

        self._ui["button_cancel"].setText(t("Cancel"))
        self._ui["button_continue"].setText("")

        self._continue = lambda: None

        self._transactions = TransactionList()
        self._model = TransactionListModel(self._transactions, self._changed)
        self._ui["table"].setModel(self._model)

        self._steps = [
            {
                "init_button": t("Collect Snapshot Transactions"),
                "prepare": self._prepare_snap,
                "prepare_text": t("Collecting snapshot transactions ..."),
                "run_text": t("Creating snapshots ..."),
                "run_button": t("Execute Snapshot Transactions"),
                "finish_text": t("Snapshots created."),
                "finish_button": t("Collect Backup Transactions"),
            },
            {
                "prepare": lambda: self._prepare("backup"),
                "prepare_text": t("Collecting backup transactions ..."),
                "run_text": t("Transferring snapshots ..."),
                "run_button": t("Execute Backup Transactions"),
                "finish_text": t("Snapshots transferred."),
                "finish_button": t("Collect Cleanup Transactions"),
            },
            {
                "prepare": lambda: self._prepare("cleanup"),
                "prepare_text": t("Collecting cleanup transactions ..."),
                "run_text": t("Removing old snapshots ..."),
                "run_button": t("Execute Cleanup Transactions"),
                "finish_text": t("Old snapshots removed."),
                "finish_button": t("Close"),
            },
        ]

        self._init_step(0)

    def _changed(self):

        self._ui["table"].resizeColumnsToContents()
        QApplication.processEvents()

    def _continue_click(self):

        self._continue()

    def _cancel_click(self):

        self._quit()

    def _init_step(self, index: int):

        self._ui["button_cancel"].setEnabled(False)
        self._ui["button_continue"].setEnabled(False)
        self._ui["progress"].setValue(0)
        self._ui["label"].setText(self._steps[index]["prepare_text"])
        self._transactions.clear()

        if index > 0:
            self._prepare_step(index)
            return

        self._continue = lambda: self._prepare_step(index)
        self._ui["button_continue"].setText(self._steps[index]["init_button"])
        self._ui["button_cancel"].setEnabled(True)
        self._ui["button_continue"].setEnabled(True)

    def _prepare_step(self, index: int):

        self._ui["button_cancel"].setEnabled(False)
        self._ui["button_continue"].setEnabled(False)
        QApplication.processEvents()

        if not self._steps[index]["prepare"]():
            QMessageBox.warning(self, t("Warning"), t("Nothing to do."))
            self._finish_step(index)
            return

        self._ui["label"].setText(self._steps[index]["run_text"])
        self._continue = lambda: self._run_step(index)
        self._ui["button_continue"].setText(self._steps[index]["run_button"])
        self._ui["button_cancel"].setEnabled(True)
        self._ui["button_continue"].setEnabled(True)

    def _run_step(self, index: int):

        self._ui["button_cancel"].setEnabled(False)
        self._ui["button_continue"].setEnabled(False)
        self._ui["progress"].setMaximum(len(self._transactions))
        QApplication.processEvents()

        for number, transaction in enumerate(self._transactions):

            assert not transaction.running
            assert not transaction.complete

            transaction.run()
            self._ui["progress"].setValue(number + 1)
            QApplication.processEvents()

            assert not transaction.running
            assert transaction.complete

            if transaction.error is not None:
                QMessageBox.critical(
                    self,
                    t("Critical Error"),
                    t("Transaction failed!")
                    + "\n\n"
                    + "\n\n".join([str(item) for item in transaction.error.args]),
                )
                self._quit()
                return

        self._ui["label"].setText(self._steps[index]["finish_text"])
        self._continue = lambda: self._finish_step(index)

        if index + 1 == len(self._steps):
            self._ui["button_cancel"].setVisible(False)
        else:
            self._ui["button_cancel"].setEnabled(True)
        self._ui["button_continue"].setText(self._steps[index]["finish_button"])
        self._ui["button_continue"].setEnabled(True)

    def _finish_step(self, index: int):

        if index + 1 == len(self._steps):
            self._quit()
            return

        self._init_step(index + 1)

    def _prepare_snap(self):

        zpool = Zpool.from_config("source", config=self._config)

        gen = zpool.generate_snapshot_transactions()
        length, _ = next(gen)

        self._ui["progress"].setMaximum(length)
        QApplication.processEvents()

        for number, transaction in gen:
            if transaction is not None:
                self._transactions.append(transaction)
            self._ui["progress"].setValue(number + 1)
            QApplication.processEvents()

        return len(self._transactions) > 0

    def _prepare(self, action: str):

        source_zpool = Zpool.from_config("source", config=self._config)
        target_zpool = Zpool.from_config("target", config=self._config)

        gen = getattr(source_zpool, f"generate_{action:s}_transactions")(target_zpool)
        length, _ = next(gen)

        self._ui["progress"].setMaximum(length)
        QApplication.processEvents()

        for number, transactions in gen:
            if transactions is not None:
                self._transactions.extend(transactions)
            self._ui["progress"].setValue(number + 1)
            QApplication.processEvents()

        return len(self._transactions) > 0

    def _quit(self):

        self._transactions.clear()
        self.close()



# -*- coding: utf-8 -*-

"""

ABGLEICH
zfs sync tool
https://github.com/pleiszenburg/abgleich

    src/abgleich/core/command.py: Sub-process wrapper for commands

    Copyright (C) 2019-2020 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the GNU Lesser General Public License
Version 2.1 ("LGPL" or "License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
https://github.com/pleiszenburg/abgleich/blob/master/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

import subprocess
import typing

import typeguard

from .abc import CommandABC

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# CLASS
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


@typeguard.typechecked
class Command(CommandABC):
    def __init__(self, cmd: typing.List[str]):

        self._cmd = cmd.copy()

    def __str__(self) -> str:

        return " ".join([item.replace(" ", "\\ ") for item in self._cmd])

    def run(
        self, returncode: bool = False
    ) -> typing.Union[typing.Tuple[str, str], typing.Tuple[str, str, int, Exception]]:

        proc = subprocess.Popen(
            self.cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
        output, errors = proc.communicate()
        status = not bool(proc.returncode)
        output, errors = output.decode("utf-8"), errors.decode("utf-8")

        exception = SystemError("command failed", str(self), output, errors)

        if returncode:
            return output, errors, int(proc.returncode), exception

        if not status or len(errors.strip()) > 0:
            raise exception

        return output, errors

    def run_pipe(self, other: CommandABC):

        proc_1 = subprocess.Popen(
            self.cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
        proc_2 = subprocess.Popen(
            other.cmd,
            stdin=proc_1.stdout,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

        output_2, errors_2 = proc_2.communicate()
        status_2 = not bool(proc_2.returncode)
        _, errors_1 = proc_1.communicate()
        status_1 = not bool(proc_1.returncode)

        errors_1 = errors_1.decode("utf-8")
        output_2, errors_2 = output_2.decode("utf-8"), errors_2.decode("utf-8")

        if any(
            (
                not status_1,
                len(errors_1.strip()) > 0,
                not status_2,
                len(errors_2.strip()) > 0,
            )
        ):
            raise SystemError(
                "command pipe failed",
                f"{str(self):s} | {str(other):s}",
                errors_1,
                output_2,
                errors_2,
            )

        return errors_1, output_2, errors_2

    @property
    def cmd(self) -> typing.List[str]:

        return self._cmd.copy()

    @classmethod
    def on_side(
        cls, cmd: typing.List[str], side: str, config: typing.Dict
    ) -> CommandABC:

        if config[side]["host"] == "localhost":
            return cls(cmd)
        return cls.with_ssh(cmd, side_config=config[side], ssh_config=config["ssh"])

    @classmethod
    def with_ssh(
        cls, cmd: typing.List[str], side_config: typing.Dict, ssh_config: typing.Dict
    ) -> CommandABC:

        cmd_str = " ".join([item.replace(" ", "\\ ") for item in cmd])
        cmd = [
            "ssh",
            "-T",  # Disable pseudo-terminal allocation
            "-o",
            "Compression=yes" if ssh_config["compression"] else "Compression=no",
        ]
        if ssh_config["cipher"] is not None:
            cmd.extend(("-c", ssh_config["cipher"]))
        cmd.extend([f'{side_config["user"]:s}@{side_config["host"]:s}', cmd_str])

        return cls(cmd)



# -*- coding: utf-8 -*-

"""

ABGLEICH
zfs sync tool
https://github.com/pleiszenburg/abgleich

	src/abgleich/core/__init__.py: Core package root

	Copyright (C) 2019-2020 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the GNU Lesser General Public License
Version 2.1 ("LGPL" or "License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
https://github.com/pleiszenburg/abgleich/blob/master/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""



# -*- coding: utf-8 -*-

"""

ABGLEICH
zfs sync tool
https://github.com/pleiszenburg/abgleich

    src/abgleich/core/zpool.py: ZFS zpool

    Copyright (C) 2019-2020 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the GNU Lesser General Public License
Version 2.1 ("LGPL" or "License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
https://github.com/pleiszenburg/abgleich/blob/master/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

from collections import OrderedDict
import typing

from tabulate import tabulate
import typeguard

from .abc import (
    ComparisonItemABC,
    ConfigABC,
    DatasetABC,
    SnapshotABC,
    TransactionABC,
    TransactionListABC,
    ZpoolABC,
)
from .command import Command
from .comparison import Comparison
from .dataset import Dataset
from .i18n import t
from .io import colorize, humanize_size
from .lib import join, root
from .property import Property
from .transaction import TransactionList

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# CLASS
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


@typeguard.typechecked
class Zpool(ZpoolABC):
    def __init__(
        self, datasets: typing.List[DatasetABC], side: str, config: ConfigABC,
    ):

        self._datasets = datasets
        self._side = side
        self._config = config

        self._root = root(config[side]["zpool"], config[side]["prefix"])

    def __eq__(self, other: ZpoolABC) -> bool:

        return self.side == other.side

    @property
    def datasets(self) -> typing.Generator[DatasetABC, None, None]:

        return (dataset for dataset in self._datasets)

    @property
    def side(self) -> str:

        return self._side

    @property
    def root(self) -> str:

        return self._root

    def get_cleanup_transactions(self, other: ZpoolABC,) -> TransactionListABC:

        assert self.side == "source"
        assert other.side == "target"

        zpool_comparison = Comparison.from_zpools(self, other)
        transactions = TransactionList()

        for dataset_item in zpool_comparison.merged:

            cleanup_transactions = self._get_cleanup_from_datasetitem(dataset_item)
            if cleanup_transactions is None:
                continue
            transactions.extend(cleanup_transactions)

        return transactions

    def generate_cleanup_transactions(
        self, other: ZpoolABC,
    ) -> typing.Generator[
        typing.Tuple[
            int,
            typing.Union[
                None, typing.Union[None, typing.Generator[TransactionABC, None, None]]
            ],
        ],
        None,
        None,
    ]:

        assert self.side == "source"
        assert other.side == "target"

        zpool_comparison = Comparison.from_zpools(self, other)

        yield len(zpool_comparison), None

        for index, dataset_item in enumerate(zpool_comparison.merged):
            yield index, self._get_cleanup_from_datasetitem(dataset_item)

    def _get_cleanup_from_datasetitem(
        self, dataset_item: ComparisonItemABC,
    ) -> typing.Union[None, typing.Generator[TransactionABC, None, None]]:

        if dataset_item.get_item().subname in self._config["ignore"]:
            return
        if dataset_item.a is None or dataset_item.b is None:
            return

        dataset_comparison = Comparison.from_datasets(dataset_item.a, dataset_item.b)
        snapshots = dataset_comparison.a_overlap_tail[: -self._config["keep_snapshots"]]

        return (snapshot.get_cleanup_transaction() for snapshot in snapshots)

    def get_backup_transactions(self, other: ZpoolABC,) -> TransactionListABC:

        assert self.side == "source"
        assert other.side == "target"

        zpool_comparison = Comparison.from_zpools(self, other)
        transactions = TransactionList()

        for dataset_item in zpool_comparison.merged:
            backup_transactions = self._get_backup_transactions_from_datasetitem(
                other, dataset_item
            )
            if backup_transactions is None:
                continue
            transactions.extend(backup_transactions)

        return transactions

    def generate_backup_transactions(
        self, other: ZpoolABC,
    ) -> typing.Generator[
        typing.Tuple[
            int,
            typing.Union[
                None, typing.Union[None, typing.Generator[TransactionABC, None, None]]
            ],
        ],
        None,
        None,
    ]:

        assert self.side == "source"
        assert other.side == "target"

        zpool_comparison = Comparison.from_zpools(self, other)

        yield len(zpool_comparison), None

        for index, dataset_item in enumerate(zpool_comparison.merged):
            yield index, self._get_backup_transactions_from_datasetitem(
                other, dataset_item
            )

    def _get_backup_transactions_from_datasetitem(
        self, other: ZpoolABC, dataset_item: ComparisonItemABC,
    ) -> typing.Union[None, typing.Generator[TransactionABC, None, None]]:

        if dataset_item.get_item().subname in self._config["ignore"]:
            return
        if dataset_item.a is None:
            return

        if dataset_item.b is None:
            snapshots = list(dataset_item.a.snapshots)
        else:
            dataset_comparison = Comparison.from_datasets(
                dataset_item.a, dataset_item.b
            )
            snapshots = dataset_comparison.a_head

        if len(snapshots) == 0:
            return

        source_dataset = (
            self.root
            if len(dataset_item.a.subname) == 0
            else join(self.root, dataset_item.a.subname)
        )
        target_dataset = (
            other.root
            if len(dataset_item.a.subname) == 0
            else join(other.root, dataset_item.a.subname)
        )

        return (
            snapshot.get_backup_transaction(source_dataset, target_dataset,)
            for snapshot in snapshots
        )

    def get_snapshot_transactions(self) -> TransactionListABC:

        assert self._side == "source"

        transactions = TransactionList()

        for dataset in self._datasets:
            transaction = self._get_snapshot_transactions_from_dataset(dataset)
            if transaction is None:
                continue
            transactions.append(dataset.get_snapshot_transaction())

        return transactions

    def generate_snapshot_transactions(
        self,
    ) -> typing.Generator[
        typing.Tuple[int, typing.Union[None, TransactionABC]], None, None
    ]:

        assert self._side == "source"

        yield len(self._datasets), None

        for index, dataset in enumerate(self._datasets):
            yield index, self._get_snapshot_transactions_from_dataset(dataset)

    def _get_snapshot_transactions_from_dataset(
        self, dataset: DatasetABC
    ) -> typing.Union[None, TransactionABC]:

        if dataset.subname in self._config["ignore"]:
            return
        if (
            dataset.get("mountpoint").value is None
            and dataset["type"].value == "filesystem"
        ):
            return
        if not dataset.changed:
            return

        return dataset.get_snapshot_transaction()

    def print_table(self):

        table = []
        for dataset in self._datasets:
            table.append(self._table_row(dataset))
            for snapshot in dataset.snapshots:
                table.append(self._table_row(snapshot))

        if len(table) == 0:
            print('(empty)')
            return

        print(
            tabulate(
                table,
                headers=(t("NAME"), t("USED"), t("REFER"), t("compressratio")),
                tablefmt="github",
                colalign=("left", "right", "right", "decimal"),
            )
        )

    @staticmethod
    def _table_row(entity: typing.Union[SnapshotABC, DatasetABC]) -> typing.List[str]:

        return [
            "- " + colorize(entity.name, "grey")
            if isinstance(entity, SnapshotABC)
            else colorize(entity.name, "white"),
            humanize_size(entity["used"].value, add_color=True),
            humanize_size(entity["referenced"].value, add_color=True),
            f'{entity["compressratio"].value:.02f}',
        ]

    def print_comparison_table(self, other: ZpoolABC):

        zpool_comparison = Comparison.from_zpools(self, other)
        table = []

        for dataset_item in zpool_comparison.merged:
            table.append(self._comparison_table_row(dataset_item))
            if dataset_item.complete:
                dataset_comparison = Comparison.from_datasets(
                    dataset_item.a, dataset_item.b
                )
            elif dataset_item.a is not None:
                dataset_comparison = Comparison.from_datasets(dataset_item.a, None)
            else:
                dataset_comparison = Comparison.from_datasets(None, dataset_item.b)
            for snapshot_item in dataset_comparison.merged:
                table.append(self._comparison_table_row(snapshot_item))

        print(
            tabulate(
                table,
                headers=[t("NAME"), t(self.side), t(other.side)],
                tablefmt="github",
            )
        )

    @staticmethod
    def _comparison_table_row(item: ComparisonItemABC) -> typing.List[str]:

        entity = item.get_item()
        name = entity.name if isinstance(entity, SnapshotABC) else entity.subname

        if item.a is not None and item.b is not None:
            a, b = colorize("X", "green"), colorize("X", "green")
        elif item.a is None and item.b is not None:
            a, b = "", colorize("X", "blue")
        elif item.a is not None and item.b is None:
            a, b = colorize("X", "red"), ""

        return [
            "- " + colorize(name, "grey")
            if isinstance(entity, SnapshotABC)
            else colorize(name, "white"),
            a,
            b,
        ]

    @staticmethod
    def available(side: str, config: ConfigABC,) -> int:

        output, _ = Command.on_side(
            [
                "zfs",
                "get",
                "available",
                "-H",
                "-p",
                root(config[side]["zpool"], config[side]["prefix"]),
            ],
            side,
            config,
        ).run()

        return Property.from_params(*output.strip().split("\t")[1:]).value

    @classmethod
    def from_config(cls, side: str, config: ConfigABC,) -> ZpoolABC:

        root_dataset = root(config[side]["zpool"], config[side]["prefix"])

        output, errors, returncode, exception = Command.on_side(
            ["zfs", "get", "all", "-r", "-H", "-p", root_dataset,], side, config,
        ).run(returncode = True)

        if returncode != 0 and 'dataset does not exist' in errors:
            return cls(datasets=[], side=side, config=config,)
        if returncode != 0:
            raise exception

        output = [
            line.split("\t") for line in output.split("\n") if len(line.strip()) > 0
        ]
        entities = OrderedDict((line[0], []) for line in output)
        for line_list in output:
            entities[line_list[0]].append(line_list[1:])

        if not config.get("include_root", True):
            entities.pop(root_dataset)
            for name in [
                snapshot
                for snapshot in entities.keys()
                if snapshot.startswith(f"{root_dataset:s}@")
            ]:
                entities.pop(name)

        datasets = [
            Dataset.from_entities(
                name,
                OrderedDict(
                    (k, v)
                    for k, v in entities.items()
                    if k == name or k.startswith(f"{name:s}@")
                ),
                side,
                config,
            )
            for name in entities.keys()
            if "@" not in name
        ]
        datasets.sort(key=lambda dataset: dataset.name)

        return cls(datasets=datasets, side=side, config=config,)



# -*- coding: utf-8 -*-

"""

ABGLEICH
zfs sync tool
https://github.com/pleiszenburg/abgleich

    src/abgleich/core/transaction.py: ZFS transactions

    Copyright (C) 2019-2020 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the GNU Lesser General Public License
Version 2.1 ("LGPL" or "License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
https://github.com/pleiszenburg/abgleich/blob/master/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

import typing

from tabulate import tabulate
import typeguard

from .abc import CommandABC, TransactionABC, TransactionListABC, TransactionMetaABC
from .i18n import t
from .io import colorize, humanize_size

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# CLASS
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


@typeguard.typechecked
class Transaction(TransactionABC):
    def __init__(
        self, meta: TransactionMetaABC, commands: typing.List[CommandABC],
    ):

        assert len(commands) in (1, 2)

        self._meta, self._commands = meta, commands

        self._complete = False
        self._running = False
        self._error = None

        self._changed = None

    @property
    def changed(self) -> typing.Union[None, typing.Callable]:

        return self._changed

    @changed.setter
    def changed(self, value: typing.Union[None, typing.Callable]):

        self._changed = value

    @property
    def complete(self) -> bool:

        return self._complete

    @property
    def commands(self) -> typing.Tuple[CommandABC]:

        return self._commands

    @property
    def error(self) -> typing.Union[Exception, None]:

        return self._error

    @property
    def meta(self) -> TransactionMetaABC:

        return self._meta

    @property
    def running(self) -> bool:

        return self._running

    def run(self):

        if self._complete:
            return

        self._running = True
        if self._changed is not None:
            self._changed()

        try:
            if len(self._commands) == 1:
                output, errors = self._commands[0].run()
            else:
                errors_1, output_2, errors_2 = self._commands[0].run_pipe(
                    self._commands[1]
                )
        except SystemError as error:
            self._error = error
        finally:
            self._running = False
            self._complete = True
            if self._changed is not None:
                self._changed()


MetaTypes = typing.Union[str, int, float]
MetaNoneTypes = typing.Union[str, int, float, None]


@typeguard.typechecked
class TransactionMeta(TransactionMetaABC):
    def __init__(self, **kwargs: MetaTypes):

        self._meta = kwargs

    def __getitem__(self, key: str) -> MetaTypes:

        return self._meta[key]

    def __len__(self) -> int:

        return len(self._meta)

    def get(self, key: str) -> MetaNoneTypes:

        return self._meta.get(key, None)

    def keys(self) -> typing.Generator[str, None, None]:

        return (key for key in self._meta.keys())


TransactionIterableTypes = typing.Union[
    typing.Generator[TransactionABC, None, None],
    typing.List[TransactionABC],
    typing.Tuple[TransactionABC],
]


@typeguard.typechecked
class TransactionList(TransactionListABC):
    def __init__(self):

        self._transactions = []
        self._changed = None

    def __len__(self) -> int:

        return len(self._transactions)

    def __getitem__(self, index: int) -> TransactionABC:

        return self._transactions[index]

    @property
    def changed(self) -> typing.Union[None, typing.Callable]:

        return self._changed

    @changed.setter
    def changed(self, value: typing.Union[None, typing.Callable]):

        self._changed = value

    @property
    def table_columns(self) -> typing.List[str]:

        headers = set()
        for transaction in self._transactions:
            keys = list(transaction.meta.keys())
            assert t("type") in keys
            headers.update(keys)
        headers = list(headers)
        headers.sort()

        if len(headers) == 0:
            return headers

        type_index = headers.index(t("type"))
        if type_index != 0:
            headers.pop(type_index)
            headers.insert(0, t("type"))

        return headers

    @property
    def table_rows(self) -> typing.List[str]:

        return [f'{t("transaction"):s} #{index:d}' for index in range(1, len(self) + 1)]

    def append(self, transaction: TransactionABC):

        self._transactions.append(transaction)
        if self._changed is not None:
            self._link_transaction(transaction)

    def extend(self, transactions: TransactionIterableTypes):

        transactions = list(transactions)
        self._transactions.extend(transactions)
        if self._changed is not None:
            for transaction in transactions:
                self._link_transaction(transaction)

    def clear(self):

        self._transactions.clear()
        self._changed()

    def _link_transaction(self, transaction: TransactionABC):

        transaction.changed = lambda: self._changed(
            self._transactions.index(transaction)
        )
        transaction.changed()

    def print_table(self):

        if len(self) == 0:
            return

        table_columns = self.table_columns
        colalign = self._table_colalign(table_columns)

        table = [
            [
                self._table_format_cell(header, transaction.meta.get(header))
                for header in table_columns
            ]
            for transaction in self._transactions
        ]

        print(
            tabulate(
                table, headers=table_columns, tablefmt="github", colalign=colalign,
            )
        )

    @staticmethod
    def _table_format_cell(header: str, value: MetaNoneTypes) -> str:

        FORMAT = {
            t("written"): lambda v: humanize_size(v, add_color=True),
        }

        return FORMAT.get(header, str)(value)

    @staticmethod
    def _table_colalign(headers: typing.List[str]) -> typing.List[str]:

        RIGHT = (t("written"),)
        DECIMAL = tuple()

        colalign = []
        for header in headers:
            if header in RIGHT:
                colalign.append("right")
            elif header in DECIMAL:
                colalign.append("decimal")
            else:
                colalign.append("left")

        return colalign

    def run(self):

        for transaction in self._transactions:

            print(
                f'({colorize(transaction.meta[t("type")], "white"):s}) '
                f'{colorize(" | ".join([str(command) for command in transaction.commands]), "yellow"):s}'
            )

            assert not transaction.running
            assert not transaction.complete

            transaction.run()

            assert not transaction.running
            assert transaction.complete

            if transaction.error is not None:
                print(colorize(t("FAILED"), "red"))
                raise transaction.error
            else:
                print(colorize(t("OK"), "green"))



# -*- coding: utf-8 -*-

"""

ABGLEICH
zfs sync tool
https://github.com/pleiszenburg/abgleich

    src/abgleich/core/lib.py: ZFS library

    Copyright (C) 2019-2020 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the GNU Lesser General Public License
Version 2.1 ("LGPL" or "License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
https://github.com/pleiszenburg/abgleich/blob/master/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

import re
import typing

import typeguard

from .abc import ConfigABC
from .command import Command

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# ROUTINES
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


@typeguard.typechecked
def is_host_up(side: str, config: ConfigABC) -> bool:

    assert side in ("source", "target")
    if config[side]["host"] == "localhost":
        return True

    _, _, returncode, _ = Command.on_side(["exit"], side, config).run(returncode=True)
    assert returncode in (0, 255)
    return returncode == 0


@typeguard.typechecked
def join(*args: str) -> str:

    if len(args) < 2:
        raise ValueError("not enough elements to join")

    args = [arg.strip("/ \t\n") for arg in args]

    if any((len(arg) == 0 for arg in args)):
        raise ValueError("can not join empty path elements")

    return "/".join(args)


@typeguard.typechecked
def root(zpool: str, prefix: typing.Union[str, None]) -> str:

    if prefix is None:
        return zpool
    return join(zpool, prefix)


_name_re = re.compile("^[A-Za-z0-9_]+$")


@typeguard.typechecked
def valid_name(name: str, min_len: int = 1) -> bool:

    assert min_len >= 0

    if len(name) < min_len:
        return False
    return bool(_name_re.match(name))



# -*- coding: utf-8 -*-

"""

ABGLEICH
zfs sync tool
https://github.com/pleiszenburg/abgleich

    src/abgleich/core/i18n.py: Translations

    Copyright (C) 2019-2020 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the GNU Lesser General Public License
Version 2.1 ("LGPL" or "License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
https://github.com/pleiszenburg/abgleich/blob/master/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

import locale
import os

import typeguard
import yaml

try:
    from yaml import CLoader as Loader
except ImportError:
    from yaml import FullLoader as Loader

try:
    from yaml import CDumper as Dumper
except ImportError:
    from yaml import Dumper

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# CLASS
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


@typeguard.typechecked
class _Lang(dict):
    def __init__(self):

        super().__init__()
        self._lang = locale.getlocale()[0].split("_")[0]
        self._path = os.path.join(
            os.path.dirname(__file__), "..", "share", "translations.yaml"
        )
        self._load()

    def __call__(self, name: str) -> str:

        assert len(name) > 0

        if int(os.environ.get("ABGLEICH_TRANSLATE", "0")) == 1:
            if name not in self.keys():
                self._add_item(name)

        return self.get(name, {}).get(self._lang, name)

    def _add_item(self, name: str):

        self[name] = {}
        self._dump()

    def _load(self):

        self.clear()

        with open(self._path, "r") as f:
            self.update(yaml.load(f.read(), Loader=Loader))

    def _dump(self):

        with open(self._path, "w") as f:
            f.write(yaml.dump(self.copy(), Dumper=Dumper, allow_unicode=True, indent=4))


# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# API
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

t = _Lang()



# -*- coding: utf-8 -*-

"""

ABGLEICH
zfs sync tool
https://github.com/pleiszenburg/abgleich

    src/abgleich/core/filesystem.py: ZFS filesystem

    Copyright (C) 2019-2020 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the GNU Lesser General Public License
Version 2.1 ("LGPL" or "License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
https://github.com/pleiszenburg/abgleich/blob/master/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

import typing

import typeguard

from .abc import PropertyABC

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# TYPING
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

PropertyTypes = typing.Union[str, int, float, None]

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# CLASS
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


@typeguard.typechecked
class Property(PropertyABC):
    def __init__(
        self, name: str, value: PropertyTypes, src: PropertyTypes,
    ):

        self._name = name
        self._value = value
        self._src = src

    @property
    def name(self) -> str:
        return self._name

    @property
    def value(self) -> PropertyTypes:
        return self._value

    @property
    def src(self) -> PropertyTypes:
        return self._src

    @classmethod
    def _convert(cls, value: str) -> PropertyTypes:

        value = value.strip()

        if value.isnumeric():
            return int(value)

        if value.strip() == "" or value == "-" or value.lower() == "none":
            return None

        try:
            return float(value)
        except ValueError:
            pass

        return value

    @classmethod
    def from_params(cls, name, value, src) -> PropertyABC:

        return cls(name=name, value=cls._convert(value), src=cls._convert(src),)



# -*- coding: utf-8 -*-

"""

ABGLEICH
zfs sync tool
https://github.com/pleiszenburg/abgleich

    src/abgleich/cli/snap.py: snap command entry point

    Copyright (C) 2019-2020 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the GNU Lesser General Public License
Version 2.1 ("LGPL" or "License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
https://github.com/pleiszenburg/abgleich/blob/master/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""


# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

import click
import sys

from ..core.config import Config
from ..core.i18n import t
from ..core.lib import is_host_up
from ..core.zpool import Zpool

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# ROUTINES
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


@click.command(short_help="create snapshots of changed datasets for backups")
@click.argument("configfile", type=click.File("r", encoding="utf-8"))
def snap(configfile):

    config = Config.from_fd(configfile)

    if not is_host_up("source", config):
        print(f'{t("host is not up"):s}: source')
        sys.exit(1)

    zpool = Zpool.from_config("source", config=config)
    transactions = zpool.get_snapshot_transactions()

    if len(transactions) == 0:
        print(t("nothing to do"))
        return
    transactions.print_table()

    click.confirm(t("Do you want to continue?"), abort=True)

    transactions.run()



# -*- coding: utf-8 -*-

"""

ABGLEICH
zfs sync tool
https://github.com/pleiszenburg/abgleich

    setup.py: Used for package distribution

    Copyright (C) 2019-2020 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the GNU Lesser General Public License
Version 2.1 ("LGPL" or "License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
https://github.com/pleiszenburg/abgleich/blob/master/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""


# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

from setuptools import (
    find_packages,
    setup,
)
import ast
import os

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# SETUP
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

# List all versions of Python which are supported
python_minor_min = 6
python_minor_max = 8
confirmed_python_versions = [
    "Programming Language :: Python :: 3.{MINOR:d}".format(MINOR=minor)
    for minor in range(python_minor_min, python_minor_max + 1)
]

# Fetch readme file
with open(os.path.join(os.path.dirname(__file__), "README.md")) as f:
    long_description = f.read()

# Define source directory (path)
SRC_DIR = "src"

# Version
def get_version(code):
    tree = ast.parse(code)
    for item in tree.body:
        if not isinstance(item, ast.Assign):
            continue
        if len(item.targets) != 1:
            continue
        if item.targets[0].id != "__version__":
            continue
        return item.value.s


with open(os.path.join(SRC_DIR, "abgleich", "__init__.py"), "r", encoding="utf-8") as f:
    __version__ = get_version(f.read())

# Requirements
extras_require = {
    "dev": ["black", "python-language-server[all]", "setuptools", "twine", "wheel",],
    "gui": ["pyqt5",],
}
extras_require["all"] = list(
    {rq for target in extras_require.keys() for rq in extras_require[target]}
)

# Install package
setup(
    name="abgleich",
    packages=find_packages(SRC_DIR),
    package_dir={"": SRC_DIR},
    version=__version__,
    description="zfs sync tool",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Sebastian M. Ernst",
    author_email="ernst@pleiszenburg.de",
    url="https://github.com/pleiszenburg/abgleich",
    download_url="https://github.com/pleiszenburg/abgleich/archive/v%s.tar.gz"
    % __version__,
    license="LGPLv2",
    keywords=["zfs", "ssh",],
    scripts=[],
    include_package_data=True,
    python_requires=">=3.{MINOR:d}".format(MINOR=python_minor_min),
    setup_requires=[],
    install_requires=["click", "tabulate", "pyyaml", "typeguard",],
    extras_require=extras_require,
    zip_safe=False,
    entry_points={"console_scripts": ["abgleich = abgleich.cli:cli",],},
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Environment :: X11 Applications",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Information Technology",
        "Intended Audience :: Science/Research",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: GNU Lesser General Public License v2 (LGPLv2)",
        "Operating System :: MacOS",
        "Operating System :: POSIX :: BSD",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3",
    ]
    + confirmed_python_versions
    + [
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: Implementation :: CPython",
        "Topic :: Scientific/Engineering",
        "Topic :: System",
        "Topic :: System :: Archiving",
        "Topic :: System :: Archiving :: Backup",
        "Topic :: System :: Archiving :: Mirroring",
        "Topic :: System :: Filesystems",
        "Topic :: System :: Systems Administration",
        "Topic :: Utilities",
    ],
)



# -*- coding: utf-8 -*-

"""

ABGLEICH
zfs sync tool
https://github.com/pleiszenburg/abgleich

    src/abgleich/core/dataset.py: ZFS dataset

    Copyright (C) 2019-2020 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the GNU Lesser General Public License
Version 2.1 ("LGPL" or "License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
https://github.com/pleiszenburg/abgleich/blob/master/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

import datetime
import typing

# Python <= 3.7.1 "fix"
try:
    from typing import OrderedDict as DictType
except ImportError:
    from typing import Dict as DictType

import typeguard

from .abc import ConfigABC, DatasetABC, PropertyABC, TransactionABC, SnapshotABC
from .command import Command
from .i18n import t
from .lib import root
from .property import Property
from .transaction import Transaction, TransactionMeta
from .snapshot import Snapshot

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# CLASS
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


@typeguard.typechecked
class Dataset(DatasetABC):
    def __init__(
        self,
        name: str,
        properties: typing.Dict[str, PropertyABC],
        snapshots: typing.List[SnapshotABC],
        side: str,
        config: ConfigABC,
    ):

        self._name = name
        self._properties = properties
        self._snapshots = snapshots
        self._side = side
        self._config = config

        self._root = root(config[side]["zpool"], config[side]["prefix"])

        assert self._name.startswith(self._root)
        self._subname = self._name[len(self._root) :].strip("/")

    def __eq__(self, other: DatasetABC) -> bool:

        return self.subname == other.subname

    def __len__(self) -> int:

        return len(self._snapshots)

    def __getitem__(self, key: typing.Union[str, int, slice]) -> PropertyABC:

        if isinstance(key, str):
            return self._properties[key]
        return self._snapshots[key]

    def get(
        self,
        key: typing.Union[str, int, slice],
        default: typing.Union[None, PropertyABC] = None,
    ) -> typing.Union[None, PropertyABC]:

        if isinstance(key, str):
            return self._properties.get(
                key, Property(key, None, None) if default is None else default,
            )

        assert isinstance(key, int) or isinstance(key, slice)

        try:
            return self._snapshots[key]
        except IndexError:
            return default

    @property
    def changed(self) -> bool:

        if len(self) == 0:
            return True
        if self._config["always_changed"]:
            return True
        if self._properties["written"].value == 0:
            return False
        if self._properties["type"].value == "volume":
            return True

        if self._config["written_threshold"] is not None:
            if self._properties["written"].value > self._config["written_threshold"]:
                return True

        if not self._config["check_diff"]:
            return True

        output, _ = Command.on_side(
            ["zfs", "diff", f"{self._name:s}@{self._snapshots[-1].name:s}"],
            self._side,
            self._config,
        ).run()
        return len(output.strip(" \t\n")) > 0

    @property
    def name(self) -> str:

        return self._name

    @property
    def subname(self) -> str:

        return self._subname

    @property
    def snapshots(self) -> typing.Generator[SnapshotABC, None, None]:

        return (snapshot for snapshot in self._snapshots)

    @property
    def root(self) -> str:

        return self._root

    def get_snapshot_transaction(self) -> TransactionABC:

        snapshot_name = self._new_snapshot_name()

        return Transaction(
            TransactionMeta(
                **{
                    t("type"): t("snapshot"),
                    t("dataset_subname"): self._subname,
                    t("snapshot_name"): snapshot_name,
                    t("written"): self._properties["written"].value,
                }
            ),
            [
                Command.on_side(
                    ["zfs", "snapshot", f"{self._name:s}@{snapshot_name:s}"],
                    self._side,
                    self._config,
                )
            ],
        )

    def _new_snapshot_name(self) -> str:

        today = datetime.datetime.now().strftime("%Y%m%d")
        max_snapshots = (10 ** self._config["digits"]) - 1
        suffix = self._config["suffix"] if self._config["suffix"] is not None else ""

        todays_names = [
            snapshot.name
            for snapshot in self._snapshots
            if all(
                (
                    snapshot.name.startswith(today),
                    snapshot.name.endswith(suffix),
                    len(snapshot.name)
                    == len(today) + self._config["digits"] + len(suffix),
                )
            )
        ]
        todays_numbers = [
            int(name[len(today) : len(today) + self._config["digits"]])
            for name in todays_names
            if name[len(today) : len(today) + self._config["digits"]].isnumeric()
        ]
        if len(todays_numbers) != 0:
            todays_numbers.sort()
            new_number = todays_numbers[-1] + 1
            if new_number > max_snapshots:
                raise ValueError(f"more than {max_snapshots:d} snapshots per day")
        else:
            new_number = 1

        return f"{today:s}{new_number:02d}{suffix}"

    @classmethod
    def from_entities(
        cls,
        name: str,
        entities: DictType[str, typing.List[typing.List[str]]],
        side: str,
        config: ConfigABC,
    ) -> DatasetABC:

        properties = {
            property.name: property
            for property in (Property.from_params(*params) for params in entities[name])
        }
        entities.pop(name)

        snapshots = []
        snapshots.extend(
            (
                Snapshot.from_entity(
                    snapshot_name, entities[snapshot_name], snapshots, side, config,
                )
                for snapshot_name in entities.keys()
            )
        )

        return cls(
            name=name,
            properties=properties,
            snapshots=snapshots,
            side=side,
            config=config,
        )



# -*- coding: utf-8 -*-

"""

ABGLEICH
zfs sync tool
https://github.com/pleiszenburg/abgleich

    src/abgleich/core/config.py: Handles configuration data

    Copyright (C) 2019-2020 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the GNU Lesser General Public License
Version 2.1 ("LGPL" or "License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
https://github.com/pleiszenburg/abgleich/blob/master/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""


# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

import typing

import typeguard
import yaml

try:
    from yaml import CLoader as Loader
except ImportError:
    from yaml import FullLoader as Loader

from .abc import ConfigABC
from .lib import valid_name

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# CLASS
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


@typeguard.typechecked
class Config(ConfigABC, dict):
    @classmethod
    def from_fd(cls, fd: typing.TextIO):

        ssh_schema = {
            "compression": lambda v: isinstance(v, bool),
            "cipher": lambda v: isinstance(v, str) or v is None,
        }

        side_schema = {
            "zpool": lambda v: isinstance(v, str) and len(v) > 0,
            "prefix": lambda v: isinstance(v, str) or v is None,
            "host": lambda v: isinstance(v, str) and len(v) > 0,
            "user": lambda v: isinstance(v, str) or v is None,
        }

        root_schema = {
            "source": lambda v: cls._validate(data=v, schema=side_schema),
            "target": lambda v: cls._validate(data=v, schema=side_schema),
            "keep_snapshots": lambda v: isinstance(v, int) and v >= 1,
            "suffix": lambda v: v is None or (isinstance(v, str) and valid_name(v)),
            "digits": lambda v: isinstance(v, int) and v >= 1,
            "ignore": lambda v: isinstance(v, list)
            and all((isinstance(item, str) and len(item) > 0 for item in v)),
            "ssh": lambda v: cls._validate(data=v, schema=ssh_schema),
        }

        config = yaml.load(fd.read(), Loader=Loader)
        cls._validate(data=config, schema=root_schema)
        return cls(config)

    @classmethod
    def _validate(cls, data: typing.Dict, schema: typing.Dict):

        for field, validator in schema.items():
            if field not in data.keys():
                raise KeyError(f'missing configuration field "{field:s}"')
            if not validator(data[field]):
                raise ValueError(f'invalid value in field "{field:s}"')

        return True
